package steiner;

/**
 * Binary min heap with elements being pairs of ID and key.
 * IDs are integers from 0 to n-1, with n being specified when constructing the heap.
 * Keys are real numbers.
 * 
 */
public class Heap {

	private int h[];
	private int size;
	private int pos[];
	private double keys[];
	
	/**
	 * Create a heap for IDs from 0 to n-1
	 * @param n the range of the IDs
	 */
	public Heap(int n) {
		pos = new int[n];
		keys = new double[n];
		size = 0;
		h = new int[n+1];
	}
	
	/**
	 * Reset the heap to empty
	 */
	public void reset() {
		size = 0;
	}
	
	/**
	 * Get current number of elements of the heap
	 * @return number of elements of the heap
	 */
	public int getSize() {
		return size;
	}
	
	/**
	 * Find the minimum in the heap but does not change the heap
	 * @return ID of the minimum element
	 */
	public int getMin() {
		return h[1];
	}
	
	private void bubbleDown(int i) {
		int tmp = h[i];
		while (i<<1 <= size) {
			int child = i << 1;
			
			if (child < size && keys[h[child]] > keys[h[child+1]]) 
				child++;
			
			if (keys[tmp] > keys[h[child]]) {
				h[i] = h[child];
				pos[h[i]] = i;
				i = child;
			} else break;
		}
		h[i] = tmp;
		pos[tmp] = i;
	}
	
	private void bubbleUp(int i) {
		int tmp = h[i];
		while (i > 1) {
			int parent = i / 2;
			if (keys[tmp] < keys[h[parent]]) {
				h[i] = h[parent];
				pos[h[i]] = i;
				i = parent;
			} else break;
		}
		h[i] = tmp;
		pos[tmp] = i;
	}
	
	/**
	 * Remove the minimum element from the heap
	 * @return the ID of the element with minimum key
	 */
	public int extractMin() {
		int tmp = h[1];
		h[1] = h[size--];
		bubbleDown(1);
		return tmp;
	}
	
	/**
	 * 
	 * @param id
	 * @return Key of element id
	 */
	public double getKey(int id) {
		return keys[id];
	}
	
	/**
	 * Insert new element with given ID and key
	 * @param id
	 * @param key
	 */
	public void insert(int id, double key) {
		keys[id] = key;
		size++;
		h[size] = id;
		pos[id] = size;
		bubbleUp(size);
	}
	
	/**
	 * Decrease the key of element id to the new value
	 * @param id ID of the element being changed
	 * @param newKey new value for the key
	 */
	public void decreaseKey(int id, double newKey) {
		keys[id] = newKey;
		bubbleUp(pos[id]);
	}
}
