package steiner;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.util.Scanner;

/**
 *  Calculate approximate Steiner trees via shortest path distances and minimum spanning
 *  trees.
 */

public class Steiner {
	
	/**
	 *  number of vertices
	 */
	private static int n;
	
	/**
	 *  number of terminals
	 */
	private static int k;
	
	/**
	 * out-degrees of the vertices
	 */
	private static int deg[];
	
	/**
	 * adjacency lists of the graph.
	 * neighbors[i] is an array listing the neighbors of vertex i
	 */
	private static int neighbors[][];
	
	/**
	 * weights of the edges of the graph.
	 * w[i][j] is the weight of the j-th out-going edge from vertex i to vertex
	 * neighbor[i][j]
	 */
	private static double w[][];
	
	// longitudes and latitudes of street intersections 
	private static String lon[];
	private static String lat[];
	private static double dlon[];
	private static double dlat[];
	
	/**
	 * IDs of the terminals (numbers between 0 and n-1)
	 */
	private static int locs[];
	
	/**
	 * read the graph from graph.txt and construct the
	 * adjacency lists
	 */
	private static void readGraph() {
		try {
			Scanner sc = new Scanner(new File("graph.txt"));
			n = sc.nextInt();
			sc.nextInt(); // skip the number of edges
			lon = new String[n];
			lat = new String[n];
			dlon = new double[n];
			dlat = new double[n];
			for (int i = 0; i < n; i++) {
				lon[i] = sc.next();
				dlon[i] = Double.parseDouble(lon[i]);
				lat[i] = sc.next();
				dlat[i] = Double.parseDouble(lat[i]);
			}
			neighbors = new int[n][];
			w = new double[n][];
			deg = new int[n];
			for (int i = 0; i < n; i++) {
				deg[i] = sc.nextInt();
				neighbors[i] = new int[deg[i]];
				w[i] = new double[deg[i]];
				for (int j = 0; j < deg[i]; j++) {
					neighbors[i][j] = sc.nextInt();
					w[i][j] = sc.nextDouble();
				}
			}
			sc.close();
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Compute distance between two locations from their longitudes and latitudes
	 * @param lon1 longitude of the first location
	 * @param lat1 latitude of the first location
	 * @param lon2 longitude of the second location
	 * @param lat2 latitude of the second location
	 * @return distance between the two locations
	 */
	private static double computeDistance(double lon1, double lat1, double lon2, double lat2) {
		double phi1 = lat1 * Math.PI / 180;
		double phi2 = lat2 * Math.PI / 180;
		double diffphi = (lat2-lat1) * Math.PI / 180;
		double difflambda = (lon2-lon1) * Math.PI /180;

		double a = Math.sin(diffphi/2) * Math.sin(diffphi/2) +
		        Math.cos(phi1) * Math.cos(phi2) *
		        Math.sin(difflambda/2) * Math.sin(difflambda/2);
		return 2 * Math.atan2(Math.sqrt(a), Math.sqrt(1-a));
	}
	
	/**
	 * Read the desirable terminals and map them to the nearest 
	 * vertex in the graph
	 */
	private static void readLocs() {
		try {
			Scanner sc = new Scanner(new File("locations.txt"));
			k = sc.nextInt();
			locs = new int[k];
			for (int i = 0; i < k; i++) {
				double longitude = sc.nextDouble();
				double latitude = sc.nextDouble();
				sc.nextLine();
				double min_dist = Double.MAX_VALUE;
				int closest = 0;
				for (int j = 0; j < n; j++) {
					double tmp = computeDistance(longitude, latitude, dlon[j], dlat[j]);
					if (tmp < min_dist) {
						min_dist = tmp;
						closest = j;
					}
				}
				locs[i] = closest;
			}
			sc.close();
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Write the computed shortest path to a KML file.
	 * @param n the number of vertices
	 * @param pred the predecessor arrays for the shortest paths. pred[i] is the vertex
	 * right before i on the shortest path from the source to i.
	 * @param source ID of the source vertex
	 * @param dest ID of the destination vertex
	 * @param name name of the KML file
	 */	
	private static void savePath(int n, int pred[], int source, int dest, String name) {
		try {
			PrintWriter writer = new PrintWriter(name);
			writer.println("<?xml version=\"1.0\" encoding=\"utf-8\" ?>");
			writer.println("<kml xmlns=\"http://www.opengis.net/kml/2.2\">");
			writer.println("<Document id=\"root_doc\">");
			writer.println("<Folder><name>tree</name>");
			for (int i = dest; i != source; i = pred[i]) {
				writer.println("<Placemark><Style><LineStyle><color>ff0000ff</color></LineStyle><PolyStyle><fill>0</fill></PolyStyle></Style>");
				writer.println("<LineString><coordinates>");
				String s = lon[pred[i]]+","+lat[pred[i]]+" "+lon[i]+","+lat[i];
				writer.println(s);
				writer.println("</coordinates></LineString></Placemark>");
			}
			writer.println("</Folder></Document></kml>");
			writer.close();
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	/**
	 * Write the computed tree to a KML file.
	 * @param n the number of vertices
	 * @param pred the predecessor arrays for the shortest paths. pred[i] is the 
	 * predecessor array for the source being the i-th terminal.
	 * @param k the number of terminals we connect using the tree
	 * @param locs IDs of the terminals being connected
	 * @param tree minimum spanning tree of the complete graph on the terminals.
	 * The tree edges are the pairs (pred[i], i) except for exactly one node for which
	 * pred[i] = i.
	 * @param name name of the KML file
	 */
	private static void saveTree(int n, int pred[][], int k, int locs[], int tree[], String name) {
		try {
			PrintWriter writer = new PrintWriter(name);
			writer.println("<?xml version=\"1.0\" encoding=\"utf-8\" ?>");
			writer.println("<kml xmlns=\"http://www.opengis.net/kml/2.2\">");
			writer.println("<Document id=\"root_doc\">");
			writer.println("<Folder><name>tree</name>");
			for (int j = 0; j < k; j++) 
			if (tree[j] != j) {
				int source = locs[tree[j]];
				int dest = locs[j];
				for (int i = dest; i != source; i = pred[tree[j]][i]) {
					writer.println("<Placemark><Style><LineStyle><color>ff0000ff</color></LineStyle><PolyStyle><fill>0</fill></PolyStyle></Style>");
					writer.println("<LineString><coordinates>");
					String s = lon[pred[tree[j]][i]]+","+lat[pred[tree[j]][i]]+" "+lon[i]+","+lat[i];
					writer.println(s);
					writer.println("</coordinates></LineString></Placemark>");
				}
			}
			writer.println("</Folder></Document></kml>");
			writer.close();
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
	}

	public static void main(String[] args) {
		readGraph();
		readLocs();
		
		// compute distance from all terminals to all vertices in the graph
		// and store the distances and predecessors in d[][] and pred[][]
		double d[][] = new double[k][n];
		int pred[][] = new int[k][n];
		
		// USE YOUR DIJKSTRA CLASS TO COMPUTE SHORTEST PATH DISTANCE 
		
		// compute minimum spanning tree among the terminals using distances you just computed
		int tree[] = new int[k];
		
		// USE YOUR PRIM CLASS TO COMPUTE MINIMUM SPANNING TREES
	
		// save computed tree to tree.kml for viewing
		saveTree(n, pred, k, locs, tree, "tree.kml");
	}

}
